# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Transformation and aggregation functions for the aggregate module."""

from abc import ABC, abstractmethod
from collections import Counter
from typing import Any, Iterable, Optional

from metldata.builtin_transformations.aggregate.models import (
    MinimalClass,
    MinimalNamedSlot,
)
from metldata.transform.base import MetadataTransformationError

_FUNCTION_REGISTRY = {}


class AggregationFunction(ABC):
    """An abstract class for aggregation transformation functions"""

    result_range_name: str
    """The name of the range of the data produced by func."""

    result_range_cls_def: Optional[MinimalClass]
    """The class definition of the data produced by func. None if the function
    result range is a type rather than a class."""

    result_multivalued: bool
    """Whether or not the transformation function produces a single value or a
    list."""

    @classmethod
    @abstractmethod
    def func(cls, data: Iterable[Any]) -> Any:
        """Transforms input data."""


def register_function(func: type[AggregationFunction]) -> type[AggregationFunction]:
    """Registers a function in the aggregation function registry."""
    _FUNCTION_REGISTRY[func.__name__] = func
    return func


class CopyAggregation(AggregationFunction, ABC):
    """An abstract base class for type-specific copy aggregation functions."""

    @classmethod
    def _extract_single_value(cls, data: Iterable[Any]) -> Any:
        iterator = iter(data)
        value = next(iterator)
        try:
            next(iterator)
        except StopIteration:
            return value
        raise MetadataTransformationError(
            "Multiple values passed to copy aggregation function where only a"
            " single value was expected."
        )


@register_function
class StringListCopyAggregation(CopyAggregation):
    """Transformation that returns a list of strings."""

    result_range_name = "string"
    result_range_cls_def = None
    result_multivalued = True

    @classmethod
    def func(cls, data: Iterable[Any]) -> list[str]:
        return list[str](data)


@register_function
class StringCopyAggregation(CopyAggregation):
    """Transformation that returns a single string."""

    result_range_name = "string"
    result_range_cls_def = None
    result_multivalued = False

    @classmethod
    def func(cls, data: Iterable[Any]) -> str:
        return str(cls._extract_single_value(data=data))


@register_function
class IntegerCopyAggregation(CopyAggregation):
    """Transformation that returns a single integer value."""

    result_range_name = "integer"
    result_range_cls_def = None
    result_multivalued = False

    @classmethod
    def func(cls, data: Iterable[Any]) -> int:
        return int(cls._extract_single_value(data=data))


@register_function
class CountAggregation(AggregationFunction):
    """Transformation that returns the count of elements for a given sequence of
    values."""

    result_range_name = "integer"
    result_range_cls_def = None
    result_multivalued = False

    @classmethod
    def func(cls, data: Iterable[Any]) -> int:
        return sum(1 for _ in data)


class ElementCountAggregation(AggregationFunction, ABC):
    """Aggregation that returns the counts of unique elements in the given data."""

    result_multivalued = True


@register_function
class StringElementCountAggregation(ElementCountAggregation):
    """Aggregation that returns the counts of unique string elements in the
    given data."""

    result_range_name = "StringValueCount"
    result_range_cls_def = MinimalClass(
        {
            MinimalNamedSlot(range="string", multivalued=False, slot_name="value"),
            MinimalNamedSlot(range="integer", multivalued=False, slot_name="count"),
        }
    )

    @classmethod
    def func(cls, data: Iterable[Any]) -> list[dict[str, Any]]:
        return [
            {"value": str(value), "count": count}
            for value, count in Counter(data).items()
        ]


@register_function
class IntegerElementCountAggregation(ElementCountAggregation):
    """Aggregation that returns the counts of unique integer elements in the
    given data."""

    result_range_name = "IntegerValueCount"
    result_range_cls_def = MinimalClass(
        {
            MinimalNamedSlot(range="integer", multivalued=False, slot_name="value"),
            MinimalNamedSlot(range="integer", multivalued=False, slot_name="count"),
        }
    )

    @classmethod
    def func(cls, data: Iterable[Any]) -> list[dict[str, Any]]:
        return [
            {"value": int(value), "count": count}
            for value, count in Counter(data).items()
        ]


def transformation_by_name(name: str) -> type[AggregationFunction]:
    """Returns a transformation class type based on the transformation class name.

    Args:
        name (str): The transformation class name

    Raises:
        KeyError: If 'name' cannot be resolved to a known transformation class.

    Returns:
        type: The transformation class type
    """
    return _FUNCTION_REGISTRY[name + "Aggregation"]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for the publication of source events."""

import asyncio
import json

from hexkit.protocols.eventpub import EventPublisherProtocol

from metldata.event_handling.models import SubmissionAnnotation, SubmissionEventPayload
from metldata.event_handling.submission_events import SourceEventConfig
from metldata.submission_registry import models


class SourceEventPublisherConfig(SourceEventConfig):
    """Config parameters and their defaults."""


class SourceEventPublisher:
    """Handles publication of source events."""

    def __init__(
        self, *, config: SourceEventPublisherConfig, provider: EventPublisherProtocol
    ):
        """Initialize with config parameters."""

        self._config = config
        self._provider = provider

    def publish_submission(self, submission: models.Submission):
        """Publish the current submission as source event"""

        if submission.content is None:
            raise ValueError("Submission content must be defined.")

        payload = SubmissionEventPayload(
            submission_id=submission.id,
            content=submission.content,
            annotation=SubmissionAnnotation(accession_map=submission.accession_map),
        )

        asyncio.run(
            self._provider.publish(
                topic=self._config.source_event_topic,
                type_=self._config.source_event_type,
                key=submission.id,
                payload=json.loads(payload.json()),
            )
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the accession registry."""

from metldata.accession_registry.accession_registry import AccessionRegistry
from metldata.accession_registry.accession_store import AccessionStore
from metldata.config import SubmissionConfig
from tests.fixtures.config import config_sub_fixture  # noqa: F401


def test_get_accession_happy(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
):
    """Test the happy path of getting 10 accession for each resource type."""

    accession_store = AccessionStore(config=config_sub_fixture)
    accession_registry = AccessionRegistry(
        config=config_sub_fixture, accession_store=accession_store
    )

    accessions = []
    for resource_type in config_sub_fixture.prefix_mapping:
        for _ in range(10):
            # generate 10 accessions for each resource type
            expected_prefix = config_sub_fixture.prefix_mapping[resource_type]
            expected_length = len(expected_prefix) + config_sub_fixture.suffix_length

            accession = accession_registry.get_accession(resource_type=resource_type)

            assert accession.startswith(expected_prefix)
            assert len(accession) == expected_length

            accessions.append(accession)

    # check whether all theses accessions have been stored:
    for accession in accessions:
        assert accession_store.exists(accession=accession)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Functionality to publish and consume events stored on the file system."""

import json
from pathlib import Path
from typing import Iterator, Optional

from hexkit.base import InboundProviderBase
from hexkit.custom_types import Ascii, JsonObject
from hexkit.protocols.eventpub import EventPublisherProtocol
from hexkit.protocols.eventsub import EventSubscriberProtocol
from pydantic import BaseModel, BaseSettings, Field


class FileSystemEventConfig(BaseSettings):
    """Config paramters and their defaults."""

    event_store_path: Path = Field(
        ...,
        description=(
            "Path of the directory on the file system where all events are stored."
            + " Each topic is a sub-directory. Each event is stored as a"
            + " JSON file within the sub-directory for the topic. The file name corresponds"
            + " to the event key. The event type is stored together with the payload in"
            + " the event file."
        ),
    )


class Event(BaseModel):
    """An event."""

    topic: str
    type_: str
    key: str
    payload: JsonObject


def get_topic_path(*, topic: str, event_store_path: Path) -> Path:
    """Get the path of an event topic."""

    return event_store_path / topic


def get_event_path(*, topic: str, key: str, event_store_path: Path) -> Path:
    """Get the path of an event file."""

    return (
        get_topic_path(topic=topic, event_store_path=event_store_path) / f"{key}.json"
    )


def write_event(*, event: Event, event_store_path: Path) -> None:
    """Write an event to the file system."""

    event_path = get_event_path(
        topic=event.topic, key=event.key, event_store_path=event_store_path
    )
    event_content = {"type_": event.type_, "payload": event.payload}

    event_path.parent.mkdir(parents=True, exist_ok=True)
    with open(event_path, "w", encoding="utf-8") as event_file:
        json.dump(event_content, event_file, indent=4)


def read_event_file(event_path: Path) -> Event:
    """Read an event from a file."""

    with open(event_path, "r", encoding="utf-8") as event_file:
        event_content = json.load(event_file)

    return Event(
        topic=event_path.parent.name,
        key=event_path.stem,
        **event_content,
    )


def read_events_from_topic(*, topic: str, event_store_path: Path) -> Iterator[Event]:
    """Read all events for the given topic from the file system."""

    topic_path = get_topic_path(topic=topic, event_store_path=event_store_path)

    topic_path.mkdir(parents=True, exist_ok=True)

    for event_path in sorted(topic_path.iterdir()):
        yield read_event_file(event_path)


class FileSystemEventPublisher(EventPublisherProtocol):
    """An EventPublisher that stores events on the file system.

    Please note that this file system based event store mimics the behaviour of
    compacted topics. Only the last event with a given key is stored."""

    def __init__(self, config: FileSystemEventConfig):
        """Initialize with config."""

        self._config = config

    async def _publish_validated(
        self, *, payload: JsonObject, type_: Ascii, key: Ascii, topic: Ascii
    ) -> None:
        """Publish an event with already validated topic and type.

        Args:
            payload: The payload to ship with the event.
            type_: The event type. ASCII characters only.
            key: The event key. ASCII characters only.
            topic: The event topic. ASCII characters only.
        """

        event = Event(
            topic=topic,
            type_=type_,
            key=key,
            payload=payload,
        )
        write_event(
            event=event,
            event_store_path=self._config.event_store_path,
        )


class FileSystemEventSubscriber(InboundProviderBase):
    """An EventSubscriber that reads events from the file system."""

    def __init__(
        self, *, config: FileSystemEventConfig, translator: EventSubscriberProtocol
    ):
        """Initialize with config."""

        self._config = config
        self._translator = translator

        if len(self._translator.topics_of_interest) != 1:
            raise ValueError(
                "FileSystemEventSubscriber only supports a single topic of interest."
            )

        self._topic_of_interest = self._translator.topics_of_interest.pop()

    async def run(self, forever: bool = True) -> None:
        """
        Runs the inbound provider.

        Please note, unlike other inbound providers, it exits when no further events
        are available.
        """

        if not forever:
            raise NotImplementedError

        for event in read_events_from_topic(
            topic=self._topic_of_interest,
            event_store_path=self._config.event_store_path,
        ):
            await self._translator.consume(
                payload=event.payload, type_=event.type_, topic=event.topic
            )


class FileSystemEventCollector:
    """An alternative to the FileSystemEventSubscriber that can be used to obtain a
    collection of events without the need to hand over individual events to a
    translator."""

    def __init__(self, *, config: FileSystemEventConfig):
        """Initialize with config."""

        self._config = config

    def collect_events(
        self, *, topic: str, types: Optional[list[str]] = None
    ) -> Iterator[Event]:
        """Collect all events for the given types from the given topic."""

        events = read_events_from_topic(
            topic=topic,
            event_store_path=self._config.event_store_path,
        )

        for event in events:
            if not types or event.type_ in types:
                yield event



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the models module."""

import pytest
from pydantic import ValidationError

from metldata.builtin_transformations.merge_slots.models import SlotMergeInstruction


def test_slot_merge_instruction_overlap():
    """Test that an overlap in source and target slots fails with the expected
    exception."""

    with pytest.raises(ValidationError):
        SlotMergeInstruction(
            class_name="class_a",
            source_slots=["some_slot", "another_slot"],
            target_slot="some_slot",
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test reference utils."""

from metldata.builtin_transformations.infer_references.config import (
    ReferenceInferenceConfig,
)
from metldata.builtin_transformations.infer_references.path.path import ReferencePath
from metldata.builtin_transformations.infer_references.reference import (
    InferredReference,
)


def test_config():
    """Test the ReferenceInferenceConfig class."""

    inferred_ref_map = {
        "class_a": {
            "class_d": {
                "path": "class_a(class_b)>class_b(class_d)>class_d",
                "multivalued": False,
            },
            "class_c": {
                "path": "class_a(class_b)>class_b<(class_c)class_c",
                "multivalued": True,
            },
        },
        "class_b": {
            "class_c": {
                "path": "class_b<(class_c)class_c",
                "multivalued": True,
            }
        },
    }
    expected_refs = [
        InferredReference(
            source="class_a",
            target="class_d",
            path=ReferencePath(path_str="class_a(class_b)>class_b(class_d)>class_d"),
            new_slot="class_d",
            multivalued=False,
        ),
        InferredReference(
            source="class_a",
            target="class_c",
            path=ReferencePath(path_str="class_a(class_b)>class_b<(class_c)class_c"),
            new_slot="class_c",
            multivalued=True,
        ),
        InferredReference(
            source="class_b",
            target="class_c",
            path=ReferencePath(path_str="class_b<(class_c)class_c"),
            new_slot="class_c",
            multivalued=True,
        ),
    ]

    config = ReferenceInferenceConfig(inferred_ref_map=inferred_ref_map)
    observed_refs = config.inferred_references
    assert expected_refs == observed_refs



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""DAO for managing artifacts."""

from hexkit.protocols.dao import DaoFactoryProtocol, DaoNaturalId
from typing_extensions import TypeAlias

from metldata.artifacts_rest.artifact_info import get_artifact_info_dict
from metldata.artifacts_rest.models import ArtifactInfo, ArtifactResource

ArtifactResourceDao: TypeAlias = DaoNaturalId[ArtifactResource]


class DaoNotFoundError(RuntimeError):
    """Raised when a DAO is not found."""

    def __init__(self, artifact_name: str, class_name: str):
        super().__init__(
            f"Could not find DAO for artifact '{artifact_name}' and class '{class_name}'."
        )


class ArtifactDaoCollection:
    """A collection of DAOs for managing artifacts and their resources."""

    @classmethod
    async def construct(
        cls,
        dao_factory: DaoFactoryProtocol,
        artifact_infos: list[ArtifactInfo],
    ):
        """Initialize the collection of DAOs."""

        artifact_info_dict = get_artifact_info_dict(artifact_infos=artifact_infos)

        # Cannot use comprehensions because of bug: https://bugs.python.org/issue33346
        artifact_daos: dict[str, dict[str, ArtifactResourceDao]] = {}
        for artifact_name, artifact_info in artifact_info_dict.items():
            artifact_daos[artifact_name] = {}
            for class_name in artifact_info.resource_classes:
                artifact_daos[artifact_name][class_name] = await dao_factory.get_dao(
                    name=cls._get_dao_name(
                        artifact_name=artifact_name, class_name=class_name
                    ),
                    dto_model=ArtifactResource,
                    id_field="id_",
                )

        return cls(artifact_daos=artifact_daos)

    def __init__(self, artifact_daos: dict[str, dict[str, ArtifactResourceDao]]):
        """Initialize the collection of DAOs.

        Please note, don't use this constructor directly. Use the construct method
        instead.
        """

        self._artifact_daos = artifact_daos

    @staticmethod
    def _get_dao_name(*, artifact_name: str, class_name: str) -> str:
        """The dao name is a combination of the name of the artifact and the name of
        the class of the resource."""

        return f"art_{artifact_name}_class_{class_name}"

    async def get_all_resource_tags(self) -> set[tuple[str, str, str]]:
        """Retrieve resource tags for all artifacts currently present in the db.

        A resource tag combines artifact_name, class_name and resource ID (accession)
        into a tuple, i.e. it has the form '(artifact_name,class_name,resource_ID)'.
        This is done to flatten the nested artifact information to simplify keeping track
        of changes.
        """

        all_resource_tags: list[tuple[str, str, str]] = []
        for artifact_type, class_name_dao in self._artifact_daos.items():
            for class_name, resource_dao in class_name_dao.items():
                # empty mapping should yield all resources
                resource_ids = [
                    resource.id_ async for resource in resource_dao.find_all(mapping={})
                ]
                all_resource_tags.extend(
                    (artifact_type, class_name, resource_id)
                    for resource_id in resource_ids
                )

        return set(all_resource_tags)

    async def get_dao(
        self, *, artifact_name: str, class_name: str
    ) -> ArtifactResourceDao:
        """Get the DAO for the given class in the given artifact.

        Raises:
            DaoNotFoundError: If the DAO could not be found.
        """

        try:
            return self._artifact_daos[artifact_name][class_name]
        except KeyError as error:
            raise DaoNotFoundError(
                artifact_name=artifact_name, class_name=class_name
            ) from error
